import _objectSpread from '@babel/runtime/helpers/esm/objectSpread2';
import _objectWithoutProperties from '@babel/runtime/helpers/esm/objectWithoutProperties';
import path from 'path';
import { ApolloError } from 'apollo-server-errors';
import { jsonFieldTypePolyfilledForSQLite } from '@k6js/ks-next/types';
import { graphql } from '@k6js/ks-next';
import _defineProperty from '@babel/runtime/helpers/esm/defineProperty';
import { Editor, Text } from 'slate';
import { c as createDocumentEditor, a as assertNever } from './index-d478a7b4.esm.js';
import * as t from 'io-ts';
import excess from 'io-ts-excess';
import { i as isValidURL } from './api-db2f9642.esm.js';
import _toPropertyKey from '@babel/runtime/helpers/esm/toPropertyKey';
import { parse, executeSync } from 'graphql';
import weakMemoize from '@emotion/weak-memoize';
import '@keystone-ui/core';
import 'react';
import 'is-hotkey';
import 'slate-react';
import 'slate-history';
import '@babel/runtime/helpers/extends';
import '@keystone-ui/popover';
import '@keystone-ui/tooltip';
import '@keystone-ui/icons/icons/LinkIcon';
import '@keystone-ui/icons/icons/Trash2Icon';
import '@keystone-ui/icons/icons/ExternalLinkIcon';
import './toolbar-01475746.esm.js';
import '@k6js/ks-next/admin-ui/context';
import '@k6js/ks-next/fields/types/relationship/views/RelationshipSelect';
import '@keystone-ui/fields';
import '@keystone-ui/button';
import '@keystone-ui/icons/icons/ColumnsIcon';
import 'apply-ref';
import '@keystone-ui/icons/icons/BoldIcon';
import '@keystone-ui/icons/icons/ItalicIcon';
import '@keystone-ui/icons/icons/PlusIcon';
import '@keystone-ui/icons/icons/ChevronDownIcon';
import '@keystone-ui/icons/icons/Maximize2Icon';
import '@keystone-ui/icons/icons/Minimize2Icon';
import '@keystone-ui/icons/icons/MoreHorizontalIcon';
import '@keystone-ui/icons/icons/CodeIcon';
import '@keystone-ui/icons/icons/AlignLeftIcon';
import '@keystone-ui/icons/icons/AlignRightIcon';
import '@keystone-ui/icons/icons/AlignCenterIcon';
import '@keystone-ui/icons/icons/MinusIcon';
import 'match-sorter';
import 'scroll-into-view-if-needed';
import 'mdast-util-from-markdown';
import 'mdast-util-gfm-autolink-literal/from-markdown';
import 'micromark-extension-gfm-autolink-literal';
import 'mdast-util-gfm-strikethrough/from-markdown';
import 'micromark-extension-gfm-strikethrough';
import 'mdast-util-definitions';
import '@braintree/sanitize-url';

// it's just about validating that it's a valid slate structure
// we'll then run normalize on it which will enforce more things

const markValue = t.union([t.undefined, t.literal(true)]);
const text = excess(t.type({
  text: t.string,
  bold: markValue,
  italic: markValue,
  underline: markValue,
  strikethrough: markValue,
  code: markValue,
  superscript: markValue,
  subscript: markValue,
  keyboard: markValue,
  insertMenu: markValue
}));

class URLType extends t.Type {
  constructor() {
    super('string', u => typeof u === 'string' && isValidURL(u), (u, c) => this.is(u) ? t.success(u) : t.failure(u, c), t.identity);

    _defineProperty(this, "_tag", 'URLType');
  }

}

const urlType = new URLType();
const link = t.recursion('Link', () => excess(t.type({
  type: t.literal('link'),
  href: urlType,
  children
})));
const relationship = t.recursion('Relationship', () => excess(t.type({
  type: t.literal('relationship'),
  relationship: t.string,
  data: t.union([t.null, relationshipData]),
  children
})));
const inline = t.union([text, link, relationship]);
const layoutArea = t.recursion('Layout', () => excess(t.type({
  type: t.literal('layout'),
  layout: t.array(t.number),
  children
})));
const onlyChildrenElements = t.recursion('OnlyChildrenElements', () => excess(t.type({
  type: t.union([t.literal('blockquote'), t.literal('layout-area'), t.literal('code'), t.literal('divider'), t.literal('list-item'), t.literal('list-item-content'), t.literal('ordered-list'), t.literal('unordered-list')]),
  children
})));
const textAlign = t.union([t.undefined, t.literal('center'), t.literal('end')]);
const heading = t.recursion('Heading', () => excess(t.type({
  type: t.literal('heading'),
  textAlign,
  level: t.union([t.literal(1), t.literal(2), t.literal(3), t.literal(4), t.literal(5), t.literal(6)]),
  children
})));
const paragraph = t.recursion('Paragraph', () => excess(t.type({
  type: t.literal('paragraph'),
  textAlign,
  children
})));
const relationshipData = excess(t.type({
  id: t.string,
  label: t.union([t.undefined, t.string]),
  data: t.union([t.undefined, t.record(t.string, t.any)])
}));
const componentBlock = t.recursion('ComponentBlock', () => excess(t.type({
  type: t.literal('component-block'),
  component: t.string,
  props: t.record(t.string, t.any),
  children
})));
const componentProp = t.recursion('ComponentProp', () => excess(t.type({
  type: t.union([t.literal('component-inline-prop'), t.literal('component-block-prop')]),
  propPath: t.union([t.array(t.union([t.string, t.number])), t.undefined]),
  children
})));
const block = t.recursion('Element', () => t.union([layoutArea, onlyChildrenElements, heading, componentBlock, componentProp, paragraph]));
const children = t.recursion('Children', () => t.array(t.union([block, inline])));
const editorCodec = t.array(block);
function isRelationshipData(val) {
  return relationshipData.validate(val, [])._tag === 'Right';
}
function validateDocumentStructure(val) {
  const result = editorCodec.validate(val, []);

  if (result._tag === 'Left') {
    throw new Error('Invalid document structure');
  }
}

class PropValidationError extends Error {
  constructor(message, path) {
    super(message);

    _defineProperty(this, "path", void 0);

    this.path = path;
  }

}

function validateComponentBlockProps(prop, value, relationships, path) {
  if (prop.kind === 'form') {
    if (prop.validate(value)) {
      return value;
    }

    throw new PropValidationError('Invalid form prop value', path);
  }

  if (prop.kind === 'child') {
    return undefined;
  }

  if (prop.kind === 'relationship') {
    // these two checks will only fail because of things done by the solution developer, not an solution user
    // TODO: move these checks
    const relationship = relationships[prop.relationship];

    if (!relationship) {
      throw new PropValidationError(`Relationship prop specifies the relationship "${prop.relationship}" but it is not an allowed relationship`, path);
    }

    if (relationship.kind !== 'prop') {
      throw new PropValidationError(`Unexpected relationship "${prop.relationship}" of kind "${relationship.kind}" where the kind should be "prop"`, path);
    }

    if (relationship.many) {
      if (Array.isArray(value) && value.every(isRelationshipData)) {
        // yes, ts understands this completely correctly, i'm as suprised as you are
        return value.map(x => ({
          id: x.id
        }));
      } else {
        throw new PropValidationError(`Invalid relationship value`, path);
      }
    }

    if (value === null || isRelationshipData(value)) {
      return value === null ? null : {
        id: value.id
      };
    } else {
      throw new PropValidationError(`Invalid relationship value`, path);
    }
  }

  if (prop.kind === 'conditional') {
    if (typeof value !== 'object' || value === null) {
      throw new PropValidationError('Conditional value must be an object', path);
    }

    for (const key of Object.keys(value)) {
      if (key !== 'discriminant' && key !== 'value') {
        throw new PropValidationError(`Conditional value only allows keys named "discriminant" and "value", not "${key}"`, path);
      }
    }

    const discriminant = value.discriminant;
    const val = value.value; // for some reason mongo or mongoose or something is saving undefined as null
    // so we're doing this so that we avoid setting undefined on objects

    const obj = {};
    const discriminantVal = validateComponentBlockProps(prop.discriminant, discriminant, relationships, path.concat('discriminant'));

    if (discriminantVal !== undefined) {
      obj.discriminant = discriminantVal;
    }

    const conditionalFieldValue = validateComponentBlockProps(prop.values[discriminant], val, relationships, path.concat('value'));

    if (conditionalFieldValue !== undefined) {
      obj.value = conditionalFieldValue;
    }

    return obj;
  }

  if (prop.kind === 'object') {
    if (typeof value !== 'object' || value === null) {
      throw new PropValidationError('Object value must be an object', path);
    }

    const allowedKeysSet = new Set(Object.keys(prop.value));

    for (const key of Object.keys(value)) {
      if (!allowedKeysSet.has(key)) {
        throw new PropValidationError(`Key on object value "${key}" is not allowed`, path);
      }
    }

    let val = {};

    for (const key of Object.keys(prop.value)) {
      const propVal = validateComponentBlockProps(prop.value[key], value[key], relationships, path.concat(key)); // for some reason mongo or mongoose or something is saving undefined as null
      // so we're doing this so that we avoid setting undefined on objects

      if (propVal !== undefined) {
        val[key] = propVal;
      }
    }

    return val;
  }

  assertNever(prop);
}

function isText(node) {
  return Text.isText(node);
} // note that the errors thrown from here will only be exposed
// as internal server error from the graphql api in prod
// this is fine because these cases are pretty much all about
// malicious content being inserted, not valid content


function getValidatedNodeWithNormalizedComponentFormProps(node, componentBlocks, relationships) {
  if (isText(node)) {
    return node;
  }

  if (node.type === 'component-block') {
    if (componentBlocks.hasOwnProperty(node.component)) {
      const componentBlock = componentBlocks[node.component];
      node = _objectSpread(_objectSpread({}, node), {}, {
        props: validateComponentBlockProps({
          kind: 'object',
          value: componentBlock.props
        }, node.props, relationships, [])
      });
    }
  }

  if (node.type === 'relationship') {
    var _node$data;

    node = {
      type: 'relationship',
      data: ((_node$data = node.data) === null || _node$data === void 0 ? void 0 : _node$data.id) !== undefined ? {
        id: node.data.id,
        data: undefined,
        label: undefined
      } : null,
      relationship: node.relationship,
      children: node.children
    };
  }

  return _objectSpread(_objectSpread({}, node), {}, {
    children: node.children.map(x => getValidatedNodeWithNormalizedComponentFormProps(x, componentBlocks, relationships))
  });
}
function validateAndNormalizeDocument(value, documentFeatures, componentBlocks, relationships) {
  validateDocumentStructure(value);
  const children = value.map(x => getValidatedNodeWithNormalizedComponentFormProps(x, componentBlocks, relationships));
  const editor = createDocumentEditor(documentFeatures, componentBlocks, relationships, {
    current: false
  });
  editor.children = children;
  Editor.normalize(editor, {
    force: true
  });
  return editor.children;
}

const labelFieldAlias = '____document_field_relationship_item_label';
const idFieldAlias = '____document_field_relationship_item_id';
function addRelationshipData(nodes, graphQLAPI, relationships, componentBlocks, gqlNames) {
  let fetchData = async (relationshipKey, data) => {
    const relationship = relationships[relationshipKey];
    if (!relationship) return data;

    if (relationship.kind === 'prop' && relationship.many) {
      const ids = Array.isArray(data) ? data.filter(item => item.id != null).map(x => x.id) : [];

      if (ids.length) {
        const labelField = getLabelFieldsForLists(graphQLAPI.schema)[relationship.listKey];
        let val = await graphQLAPI.run({
          query: `query($ids: [ID!]!) {items:${gqlNames(relationship.listKey).listQueryName}(where: { id: { in: $ids } }) {${idFieldAlias}:id ${labelFieldAlias}:${labelField}\n${relationship.selection || ''}}}`,
          variables: {
            ids
          }
        });
        return Array.isArray(val.items) ? val.items.map(_ref => {
          let {
            [labelFieldAlias]: label,
            [idFieldAlias]: id
          } = _ref,
              data = _objectWithoutProperties(_ref, [labelFieldAlias, idFieldAlias].map(_toPropertyKey));

          return {
            id,
            label,
            data
          };
        }) : [];
      }

      return [];
    } else {
      // Single related item
      const id = data === null || data === void 0 ? void 0 : data.id;

      if (id != null) {
        const labelField = getLabelFieldsForLists(graphQLAPI.schema)[relationship.listKey]; // An exception here indicates something wrong with either the system or the
        // configuration (e.g. a bad selection field). These will surface as system
        // errors from the GraphQL field resolver.

        const val = await graphQLAPI.run({
          query: `query($id: ID!) {item:${gqlNames(relationship.listKey).itemQueryName}(where: {id:$id}) {${labelFieldAlias}:${labelField}\n${relationship.selection || ''}}}`,
          variables: {
            id
          }
        });

        if (val.item === null) {
          if (!process.env.TEST_ADAPTER) {
            // If we're unable to find the item (e.g. we have a dangling reference), or access was denied
            // then simply return { id } and leave `label` and `data` undefined.
            const r = JSON.stringify(relationship);
            console.error(`Unable to fetch relationship data: relationship: ${r}, id: ${id} `);
          }

          return {
            id
          };
        }

        return {
          id,
          label: val.item[labelFieldAlias],
          data: (() => {
            const _val$item = val.item,
                  otherData = _objectWithoutProperties(_val$item, [labelFieldAlias].map(_toPropertyKey));

            return otherData;
          })()
        };
      } else {
        return null;
      }
    }
  };

  return Promise.all(nodes.map(async node => {
    if (node.type === 'relationship') {
      return _objectSpread(_objectSpread({}, node), {}, {
        data: await fetchData(node.relationship, node.data)
      });
    }

    if (node.type === 'component-block') {
      const componentBlock = componentBlocks[node.component];

      if (componentBlock) {
        const [props, children] = await Promise.all([addRelationshipDataToComponentProps({
          kind: 'object',
          value: componentBlock.props
        }, node.props, fetchData), addRelationshipData(node.children, graphQLAPI, relationships, componentBlocks, gqlNames)]);
        return _objectSpread(_objectSpread({}, node), {}, {
          props,
          children
        });
      }
    }

    if ('children' in node && Array.isArray(node.children)) {
      return _objectSpread(_objectSpread({}, node), {}, {
        children: await addRelationshipData(node.children, graphQLAPI, relationships, componentBlocks, gqlNames)
      });
    }

    return node;
  }));
}

async function addRelationshipDataToComponentProps(prop, val, fetchData) {
  switch (prop.kind) {
    case 'child':
    case 'form':
      {
        return val;
      }

    case 'relationship':
      {
        return fetchData(prop.relationship, val);
      }

    case 'object':
      {
        return Object.fromEntries(await Promise.all(Object.keys(prop.value).map(async key => [key, await addRelationshipDataToComponentProps(prop.value[key], val[key], fetchData)])));
      }

    case 'conditional':
      {
        return {
          discriminant: val.discriminant,
          value: await addRelationshipDataToComponentProps(prop.values[val.discriminant], val.value, fetchData)
        };
      }
  }

  assertNever(prop);
}

const document$1 = parse(`
  query {
    keystone {
      adminMeta {
        lists {
          key
          labelField
        }
      }
    }
  }
`);
const getLabelFieldsForLists = weakMemoize(function getLabelFieldsForLists(schema) {
  const {
    data,
    errors
  } = executeSync({
    schema,
    document: document$1,
    contextValue: {
      isAdminUIBuildProcess: true
    }
  });

  if (errors !== null && errors !== void 0 && errors.length) {
    throw errors[0];
  }

  return Object.fromEntries(data.keystone.adminMeta.lists.map(x => [x.key, x.labelField]));
});

const _excluded = ["componentBlocks", "dividers", "formatting", "layouts", "relationships", "links"];
const views = path.join(path.dirname(__dirname), 'views');
const document = (_ref = {}) => {
  let {
    componentBlocks = {},
    dividers,
    formatting,
    layouts,
    relationships: configRelationships,
    links
  } = _ref,
      config = _objectWithoutProperties(_ref, _excluded);

  return meta => {
    const documentFeatures = normaliseDocumentFeatures({
      dividers,
      formatting,
      layouts,
      links
    });
    const relationships = normaliseRelationships(configRelationships);

    const inputResolver = data => {
      if (data === null) {
        throw new ApolloError('Input error: Document fields cannot be set to null');
      }

      if (data === undefined) {
        return data;
      }

      return validateAndNormalizeDocument(data, documentFeatures, componentBlocks, relationships);
    };

    if (config.isIndexed === 'unique') {
      throw Error("isIndexed: 'unique' is not a supported option for field type document");
    }

    return jsonFieldTypePolyfilledForSQLite(meta.provider, _objectSpread(_objectSpread({}, config), {}, {
      input: {
        create: {
          arg: graphql.arg({
            type: graphql.JSON
          }),

          resolve(val) {
            if (val === undefined) {
              val = [{
                type: 'paragraph',
                children: [{
                  text: ''
                }]
              }];
            }

            return inputResolver(val);
          }

        },
        update: {
          arg: graphql.arg({
            type: graphql.JSON
          }),
          resolve: inputResolver
        }
      },
      output: graphql.field({
        type: graphql.object()({
          name: `${meta.listKey}_${meta.fieldKey}_Document`,
          fields: {
            document: graphql.field({
              args: {
                hydrateRelationships: graphql.arg({
                  type: graphql.nonNull(graphql.Boolean),
                  defaultValue: false
                })
              },
              type: graphql.nonNull(graphql.JSON),

              resolve({
                document
              }, {
                hydrateRelationships
              }, context) {
                return hydrateRelationships ? addRelationshipData(document, context.graphql, relationships, componentBlocks, context.gqlNames) : document;
              }

            })
          }
        }),

        resolve({
          value
        }) {
          if (value === null) {
            return null;
          }

          return {
            document: value
          };
        }

      }),
      views,

      getAdminMeta() {
        return {
          relationships,
          documentFeatures,
          componentBlocksPassedOnServer: Object.keys(componentBlocks)
        };
      }

    }), {
      mode: 'required',
      default: {
        kind: 'literal',
        value: JSON.stringify([{
          type: 'paragraph',
          children: [{
            text: ''
          }]
        }])
      }
    });
  };
};

function normaliseRelationships(configRelationships) {
  const relationships = {};

  if (configRelationships) {
    Object.keys(configRelationships).forEach(key => {
      var _relationship$selecti, _relationship$selecti2;

      const relationship = configRelationships[key];
      relationships[key] = relationship.kind === 'inline' ? _objectSpread(_objectSpread({}, relationship), {}, {
        selection: (_relationship$selecti = relationship.selection) !== null && _relationship$selecti !== void 0 ? _relationship$selecti : null
      }) : _objectSpread(_objectSpread({}, relationship), {}, {
        selection: (_relationship$selecti2 = relationship.selection) !== null && _relationship$selecti2 !== void 0 ? _relationship$selecti2 : null,
        many: relationship.many || false
      });
    });
  }

  return relationships;
}

function normaliseDocumentFeatures(config) {
  var _config$formatting, _formatting$alignment, _formatting$alignment2, _formatting$blockType, _formatting$blockType2, _formatting$inlineMar, _formatting$inlineMar2, _formatting$inlineMar3, _formatting$inlineMar4, _formatting$inlineMar5, _formatting$inlineMar6, _formatting$inlineMar7, _formatting$inlineMar8, _formatting$listTypes, _formatting$listTypes2;

  const formatting = config.formatting === true ? {
    alignment: true,
    blockTypes: true,
    headingLevels: true,
    inlineMarks: true,
    listTypes: true,
    softBreaks: true
  } : (_config$formatting = config.formatting) !== null && _config$formatting !== void 0 ? _config$formatting : {};
  const documentFeatures = {
    formatting: {
      alignment: formatting.alignment === true ? {
        center: true,
        end: true
      } : {
        center: !!((_formatting$alignment = formatting.alignment) !== null && _formatting$alignment !== void 0 && _formatting$alignment.center),
        end: !!((_formatting$alignment2 = formatting.alignment) !== null && _formatting$alignment2 !== void 0 && _formatting$alignment2.end)
      },
      blockTypes: (formatting === null || formatting === void 0 ? void 0 : formatting.blockTypes) === true ? {
        blockquote: true,
        code: true
      } : {
        blockquote: !!((_formatting$blockType = formatting.blockTypes) !== null && _formatting$blockType !== void 0 && _formatting$blockType.blockquote),
        code: !!((_formatting$blockType2 = formatting.blockTypes) !== null && _formatting$blockType2 !== void 0 && _formatting$blockType2.code)
      },
      headingLevels: (formatting === null || formatting === void 0 ? void 0 : formatting.headingLevels) === true ? [1, 2, 3, 4, 5, 6] : [...new Set(formatting === null || formatting === void 0 ? void 0 : formatting.headingLevels)].sort(),
      inlineMarks: formatting.inlineMarks === true ? {
        bold: true,
        code: true,
        italic: true,
        keyboard: true,
        strikethrough: true,
        subscript: true,
        superscript: true,
        underline: true
      } : {
        bold: !!((_formatting$inlineMar = formatting.inlineMarks) !== null && _formatting$inlineMar !== void 0 && _formatting$inlineMar.bold),
        code: !!((_formatting$inlineMar2 = formatting.inlineMarks) !== null && _formatting$inlineMar2 !== void 0 && _formatting$inlineMar2.code),
        italic: !!((_formatting$inlineMar3 = formatting.inlineMarks) !== null && _formatting$inlineMar3 !== void 0 && _formatting$inlineMar3.italic),
        strikethrough: !!((_formatting$inlineMar4 = formatting.inlineMarks) !== null && _formatting$inlineMar4 !== void 0 && _formatting$inlineMar4.strikethrough),
        underline: !!((_formatting$inlineMar5 = formatting.inlineMarks) !== null && _formatting$inlineMar5 !== void 0 && _formatting$inlineMar5.underline),
        keyboard: !!((_formatting$inlineMar6 = formatting.inlineMarks) !== null && _formatting$inlineMar6 !== void 0 && _formatting$inlineMar6.keyboard),
        subscript: !!((_formatting$inlineMar7 = formatting.inlineMarks) !== null && _formatting$inlineMar7 !== void 0 && _formatting$inlineMar7.subscript),
        superscript: !!((_formatting$inlineMar8 = formatting.inlineMarks) !== null && _formatting$inlineMar8 !== void 0 && _formatting$inlineMar8.superscript)
      },
      listTypes: formatting.listTypes === true ? {
        ordered: true,
        unordered: true
      } : {
        ordered: !!((_formatting$listTypes = formatting.listTypes) !== null && _formatting$listTypes !== void 0 && _formatting$listTypes.ordered),
        unordered: !!((_formatting$listTypes2 = formatting.listTypes) !== null && _formatting$listTypes2 !== void 0 && _formatting$listTypes2.unordered)
      },
      softBreaks: !!formatting.softBreaks
    },
    links: !!config.links,
    layouts: [...new Set((config.layouts || []).map(x => JSON.stringify(x)))].map(x => JSON.parse(x)),
    dividers: !!config.dividers
  };
  return documentFeatures;
}

export { document };
